import numpy as np
from cmath import exp
from math import pi, sqrt
import math


class tbmodel:
    def __init__(self,n, N=10, hermitian = True):
        self.hermitian: bool = hermitian
        self.n: int = n
        self.N: int = N
        # 这里的（0,1）对应到array是axis=0的耦合
        # 注意数学坐标系与np.ndarray的不符，imshow与plotly.scatter的不符
        self.coupling: dict = {
            (0,0): np.zeros((n,n))*0j, 
        }


    def add_coupling(self, unit_to=(0,0), atom_from = 0, atom_to = 0, value = 0):
        if unit_to not in self.coupling.keys():
            self.coupling.update({unit_to: np.zeros((self.n,self.n))*0j})
        self.coupling.get(unit_to)[atom_from, atom_to] += value
   
    
    class hmatrix:
        '''
        哈密顿矩阵对象
        set()为了避免耦合超出边界
        '''
        def __init__(self, n, imax, jmax):
            self.matrix: np.ndarray = np.zeros((n*imax*jmax, n*imax*jmax))*0j
            self.imax: int = imax
            self.jmax: int = jmax
            self.n: int = n
        
        def set(self, sub_matrix, i, j, delta_i, delta_j):
            if (i+delta_i) >= 0 and (i+delta_i) < self.imax and \
                (j+delta_j) >= 0 and (j+delta_j) < self.jmax:
                unit_ind = i*self.jmax + j
                unit_delta_ind = (i+delta_i)*self.jmax + (j+delta_j)
                self.matrix[
                    unit_ind*self.n: unit_ind*self.n + self.n,
                    unit_delta_ind*self.n: unit_delta_ind*self.n + self.n
                ] += sub_matrix



class tbsquare(tbmodel):

    def get_H_PBC(self, kx, ky):
        H_PBC = np.zeros((self.n, self.n))*0j
        for item in self.coupling.keys():
            H_PBC += self.coupling.get(item) * exp(1j*item[0]*kx) * exp(1j*item[1]*ky)
        return H_PBC + self.hermitian * H_PBC.conj().T 
    

    def get_H_xPBC(self, kx):
        H_xPBC = self.hmatrix(self.n, 1, self.N)
        for j in range(self.N):
            for item in self.coupling.keys():
                smatrix = self.coupling.get(item) * exp(1j*item[0]*kx)
                # j 的方向为周期，不会超出边界，令delta_j=0
                H_xPBC.set(smatrix, 0, j, 0, item[1])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T
    

    def get_H_xyPBC(self, kx):
        '''
        斜切45度角的周期边界条件
        '''
        H_xPBC = self.hmatrix(self.n, 1, self.N)
        for j in range(self.N):
            for item in self.coupling.keys():
                smatrix = self.coupling.get(item) * exp(1j*item[0]*kx)
                # j 的方向为周期，不会超出边界，令delta_j=0
                H_xPBC.set(smatrix, 0, j, 0, item[1]-item[0])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T
    

    def get_H_OBC(self):
        H_OBC = self.hmatrix(self.n, self.N, self.N)
        for i in range(self.N):
            for j in range(self.N):
                for item in self.coupling.keys():
                    H_OBC.set(self.coupling.get(item), i, j, item[0], item[1])
        return H_OBC.matrix + self.hermitian * H_OBC.matrix.conj().T


    def get_site_pos(self, boundary_condition='OBC', position=[(-0.25,-0.25), (0.25,-0.25), (-0.25,0.25), (0.25, 0.25)]):
        N = self.N
        x = np.empty(0)
        y = np.empty(0)
        if boundary_condition == 'PBC':
            for item in position:
                x = np.append(x,item[0])
                y = np.append(y,item[1])
        elif boundary_condition in ['xPBC', 'xyPBC']:
            for j in range(N):
                for item in position:
                    x = np.append(x,item[0])
                    y = np.append(y,j+item[1])
        elif boundary_condition == 'OBC':
            for i in range(N):
                for j in range(N):
                    for item in position:
                        x = np.append(x,i+item[0])
                        y = np.append(y,j+item[1])
        else:
            print('boundary_condition should be one of [PBC, xPBC, OBC].')
            return
        return x, y
    




class tbhexagonal(tbmodel):
    def __init__(self,n, N=10, hermitian = True, 
        base1=(sqrt(3)/2, 1/2), base2=(-sqrt(3)/2, 1/2)
    ):
        super().__init__(n, N, hermitian)
        self.base1 = base1
        self.base2 = base2


    def get_H_PBC(self, kx, ky):
        H_PBC = np.zeros((self.n, self.n))*0j
        for item in self.coupling.keys():
            ax = item[0]*self.base1[0] + item[1]*self.base2[0]
            ay = item[0]*self.base1[1] + item[1]*self.base2[1]
            H_PBC += self.coupling.get(item) * exp(1j*ax*kx) * exp(1j*ay*ky)
        return H_PBC + self.hermitian * H_PBC.conj().T
    

    def get_H_zigzag(self, k1):
        H_xPBC = self.hmatrix(self.n, 1, self.N)
        i = 0
        for j in range(self.N):
            for item in self.coupling.keys():
                smatrix = self.coupling.get(item) * exp(1j*item[0]*k1)
                # j 的方向为周期，不会超出边界，令delta_j=0
                H_xPBC.set(smatrix, i, j, 0, item[1])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T


    def get_H_armchair(self, k1):
        H_xPBC = self.hmatrix(self.n, 1, self.N)
        for j in range(self.N):
            for item in self.coupling.keys():
                smatrix = self.coupling.get(item) * exp(1j*item[0]*sqrt(3)*k1)
                H_xPBC.set(smatrix, 0, j, 0, item[1]+item[0])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T
    

    def get_H_bearded(self, k1):
        H_xPBC = self.hmatrix(self.n, 1, self.N)
        for j in range(self.N):
            for item in self.coupling.keys():
                smatrix = self.coupling.get(item) * exp(1j*item[0]*k1)
                H_xPBC.set(smatrix, 0, j, 0, item[1]-item[0])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T


    def get_H_OBC(self):
        H_OBC = self.hmatrix(self.n, self.N, self.N)
        for i in range(self.N):
            for j in range(self.N):
                for item in self.coupling.keys():
                    H_OBC.set(self.coupling.get(item), i, j, item[0], item[1])
        return H_OBC.matrix + self.hermitian * H_OBC.matrix.conj().T


    def get_site_pos(self, boundary_condition='OBC', position=[(-sqrt(3)/6,0), (sqrt(3)/6,0)]):
        N = self.N
        x = np.empty(0)
        y = np.empty(0)
        if boundary_condition == 'PBC':
            for item in position:
                x = np.append(x,item[0])
                y = np.append(y,item[1])
        elif boundary_condition in ['zigzag', 'armchair', 'bearded']:
            for j in range(N):
                for item in position:
                    x = np.append(x,j*self.base2[0]+item[0])
                    y = np.append(y,j*self.base2[1]+item[1])
        elif boundary_condition == 'OBC':
            for i in range(N):
                for j in range(N):
                    for item in position:
                        x = np.append(x,i*self.base1[0]+j*self.base2[0]+item[0])
                        y = np.append(y,i*self.base1[1]+j*self.base2[1]+item[1])
        else:
            print('boundary_condition should be one of [PBC, xPBC, OBC].')
            return
        return x, y
    


class tbmodel3D:
    def __init__(self,n, N=10, hermitian = True):
        self.hermitian: bool = hermitian
        self.n: int = n
        self.N: int = N
        self.coupling: dict = {
            (0,0,0): np.zeros((n,n))*0j, 
        }


    def add_coupling(self, unit_to=(0,0,0), atom_from = 0, atom_to = 0, value = 0):
        if unit_to not in self.coupling.keys():
            self.coupling.update({unit_to: np.zeros((self.n,self.n))*0j})
        self.coupling.get(unit_to)[atom_from, atom_to] += value
   
    
    class hmatrix:
        '''
        哈密顿矩阵对象
        set()为了避免耦合超出边界
        '''
        def __init__(self, n, imax, jmax, lmax):
            self.matrix: np.ndarray = np.zeros((n*imax*jmax*lmax, n*imax*jmax*lmax))*0j
            self.imax: int = imax
            self.jmax: int = jmax
            self.lmax: int = lmax
            self.n: int = n
        
        def set(self, sub_matrix, i, j, l, delta_i, delta_j, delta_l):
            if (i+delta_i) >= 0 and (i+delta_i) < self.imax and \
                (j+delta_j) >= 0 and (j+delta_j) < self.jmax and \
                (l+delta_l) >= 0 and (l+delta_l) < self.lmax:
                unit_ind = i*self.jmax*self.lmax + j*self.lmax + l
                unit_delta_ind = (i+delta_i)*self.jmax*self.lmax + (j+delta_j)*self.lmax + l+delta_l
                self.matrix[
                    unit_ind*self.n: unit_ind*self.n + self.n,
                    unit_delta_ind*self.n: unit_delta_ind*self.n + self.n
                ] += sub_matrix



class tbsquare3D(tbmodel3D):

    def get_H_PBC(self, kx, ky, kz):
        H_PBC = np.zeros((self.n, self.n))*0j
        for item in self.coupling.keys():
            H_PBC += self.coupling.get(item) * exp(1j*item[0]*kx) * exp(1j*item[1]*ky) * exp(1j*item[2]*kz)
        return H_PBC + self.hermitian * H_PBC.conj().T 


    def get_H_xPBC(self, kx):
        H_xPBC = self.hmatrix(self.n, 1, self.N, self.N)
        for j in range(self.N):
            for l in range(self.N):
                for item in self.coupling.keys():
                    smatrix = self.coupling.get(item) * exp(1j*item[0]*kx)
                    H_xPBC.set(smatrix, 0, j, l, 0, item[1], item[2])
        return H_xPBC.matrix + self.hermitian * H_xPBC.matrix.conj().T
     

    def get_H_OBC(self):
        H_OBC = self.hmatrix(self.n, self.N, self.N, self.N)
        for i in range(self.N):
            for j in range(self.N):
                for l in range(self.N):
                    for item in self.coupling.keys():
                        H_OBC.set(self.coupling.get(item), i, j, l, item[0], item[1], item[2])
        return H_OBC.matrix + self.hermitian * H_OBC.matrix.conj().T


    def get_site_pos(self, boundary_condition='OBC', 
            position=[(-0.25,-0.25,-0.25), (0.25,-0.25,-0.25), (-0.25,0.25,-0.25), (0.25, 0.25,-0.25),
                      (-0.25,-0.25,0.25), (0.25,-0.25,0.25), (-0.25,0.25,0.25), (0.25, 0.25,0.25)]):
        N = self.N
        x = np.empty(0)
        y = np.empty(0)
        z = np.empty(0)
        if boundary_condition == 'PBC':
            for item in position:
                x = np.append(x,item[0])
                y = np.append(y,item[1])
                z = np.append(z,item[2])

        elif boundary_condition in ['xPBC']:
            for j in range(N):
                for l in range(N):
                    for item in position:
                        x = np.append(x,item[0])
                        y = np.append(y,j+item[1])
                        z = np.append(z,l+item[2])
        elif boundary_condition == 'OBC':
            for i in range(N):
                for j in range(N):
                    for l in range(N):
                        for item in position:
                            x = np.append(x,i+item[0])
                            y = np.append(y,j+item[1])
                            z = np.append(z,l+item[2])
        
        else:
            print('boundary_condition should be one of [PBC, xPBC, OBC].')
            return

        return x, y, z



from tbcal.tbmodel import tbsquare, tbmodel
from tbcal.tbcalculation import get_val_vec, ray_casting_method 
import numpy as np
from cmath import exp
from math import pi, sqrt

class tbsqure12(tbmodel):    
    '''
    Splice two tight-binding models together.
    '''
    def __init__(self, N, tb1, tb2, tb12, cut1, cut2=None, hermitian = True):
        '''
        tb1, tb2: tight-binding models
        cut1, cut2: the cut of the two tight-binding models.
        '''
        self.tb1 = tb1
        self.tb2 = tb2
        self.tb12 = tb12
        self.cut1 = cut1
        self.cut2 = cut2
        # 判断原胞是否在区域1内
        area1 = np.zeros(N*N)
        for i in range(N):
            for j in range(N):
                if ray_casting_method(i,j,cut1):
                    area1[i*N+j] = 1
        self.area1 = area1

        self.N = N
        self.N1 = np.sum(area1)
        self.N2 = N*N - self.N1
        self.n = self.tb1.n
        if self.tb1.n != self.tb2.n:
            print('Error: the number of atoms per unit cell in two tight-binding models are not equal.')
        self.hermitian = hermitian


    def get_H_OBC(self):
        H = self.hmatrix(self.n, self.N, self.N)
        
        for i in range(self.N):
            for j in range(self.N):
                if self.area1[i*self.N+j] == 1:
                    for item in self.tb1.coupling.keys():
                        i_new = i+item[0]
                        j_new = j+item[1]
                        try:
                            is_in_area1 = self.area1[i_new*self.N+j_new]
                        except IndexError:
                            continue
                        if is_in_area1 == 1:
                            H.set(self.tb1.coupling.get(item), i, j, item[0], item[1])
                        else:
                            try:
                                H.set(self.tb12.coupling.get(item), i, j, item[0], item[1])
                            except:
                                print(i,j,i_new,j_new,'Error1: the coupling is not defined in tb3.')
                elif self.area1[i*self.N+j] == 0:
                    for item in self.tb2.coupling.keys():
                        i_new = i+item[0]
                        j_new = j+item[1]
                        try:
                            is_in_area1 = self.area1[i_new*self.N+j_new]
                        except IndexError:
                            continue
                        if is_in_area1 == 0:
                            H.set(self.tb2.coupling.get(item), i, j, item[0], item[1])
                        else:
                            try:
                                H.set(self.tb12.coupling.get(item), i, j, item[0], item[1])
                            except:
                                print(i,j,i_new,j_new,'Error2: the coupling is not defined in tb3.')           
        return H.matrix + self.hermitian * H.matrix.conj().T


    def get_site_pos(self, boundary_condition='OBC', 
        position1=[(-0.3,-0.3), (0.3,-0.3), (-0.3,0.3), (0.3, 0.3)],
        position2=[(-0.2,-0.2), (0.2,-0.2), (-0.2,0.2), (0.2, 0.2)]):
        N = self.N
        x = np.empty(0)
        y = np.empty(0)
        if boundary_condition == 'OBC':
            for i in range(N):
                for j in range(N):
                    if self.area1[i*N+j] == 1:
                        position = position1
                    elif self.area1[i*N+j] == 0:
                        position = position2
                    for item in position:
                        x = np.append(x,i+item[0])
                        y = np.append(y,j+item[1])
        return x, y
    


class tbhexagonal12(tbsqure12):   
    '''
    cut切割按照base1, base2 坐标系而不是直角坐标系
    ''' 
    def __init__(self, N, tb1, tb2, tb12, cut1, cut2=None, hermitian = True,
                 base1=(sqrt(3)/2, 1/2), base2=(-sqrt(3)/2, 1/2)):
        super().__init__(N, tb1, tb2, tb12, cut1, cut2, hermitian)
        self.base1 = base1
        self.base2 = base2


    def get_site_pos(self, boundary_condition='OBC', 
        position1=[(exp(-1j*i*pi/3+1j*pi/2).real/3*1.1,exp(-1j*i*pi/3+1j*pi/2).imag/3*1.2) for i in range(6)],
        position2=[(exp(-1j*i*pi/3+1j*pi/2).real/3/1.1,exp(-1j*i*pi/3+1j*pi/2).imag/3/1.2) for i in range(6)]):
        N = self.N
        x = np.empty(0)
        y = np.empty(0)
        if boundary_condition == 'OBC':
            for i in range(N):
                for j in range(N):
                    if self.area1[i*N+j] == 1:
                        position = position1
                    elif self.area1[i*N+j] == 0:
                        position = position2
                    for item in position:
                        x = np.append(x,i*self.base1[0]+j*self.base2[0]+item[0].real)
                        y = np.append(y,i*self.base1[1]+j*self.base2[1]+item[1].real)
        return x.real, y.real





    



import numpy as np
from cmath import exp
from math import pi, sqrt
import math



def high_symmetry_line(
        HSP: np.ndarray = np.array([[pi,pi], [pi,0], [0,0], [pi,pi]]), 
        step=0.1
    ):
    '''
    每隔 step 步长在HSP二维空间点之间插入一个点
    '''
    kx_array = np.array([HSP[0][0]])
    ky_array = np.array([HSP[0][1]])
    for i in range(HSP.shape[0]-1):
        x = HSP[i+1][0] - HSP[i][0]
        y = HSP[i+1][1] - HSP[i][1]
        k_len= sqrt(x*x + y*y)
        k_number = math.floor(k_len/step) + 1
        kx = np.linspace(HSP[i][0], HSP[i+1][0], k_number)
        kx_array = np.concatenate((kx_array, kx[1:]), axis=0)
        ky = np.linspace(HSP[i][1], HSP[i+1][1], k_number)
        ky_array = np.concatenate((ky_array, ky[1:]), axis=0)
    return kx_array, ky_array



def high_symmetry_line_3D(
        HSP: np.ndarray = np.array([[0,0,0], [pi,0,0], [pi,pi,0], [pi,pi,pi], [0,0,0]]), 
        step=0.1
    ):
    '''
    每隔 step 步长在HSP二维空间点之间插入一个点
    '''
    kx_array = np.array([HSP[0][0]])
    ky_array = np.array([HSP[0][1]])
    kz_array = np.array([HSP[0][2]])
    for i in range(HSP.shape[0]-1):
        x = HSP[i+1][0] - HSP[i][0]
        y = HSP[i+1][1] - HSP[i][1]
        z = HSP[i+1][2] - HSP[i][2]
        k_len= sqrt(x*x + y*y + z*z)
        k_number = math.floor(k_len/step) + 1
        kx = np.linspace(HSP[i][0], HSP[i+1][0], k_number)
        kx_array = np.concatenate((kx_array, kx[1:]), axis=0)
        ky = np.linspace(HSP[i][1], HSP[i+1][1], k_number)
        ky_array = np.concatenate((ky_array, ky[1:]), axis=0)
        kz = np.linspace(HSP[i][2], HSP[i+1][2], k_number)
        kz_array = np.concatenate((kz_array, kz[1:]), axis=0)
    return kx_array, ky_array, kz_array



def get_val_vec(func, data=None):
    '''
    当func为函数时，data传入列表，元素个数为func参数个数，
    每个元素为可迭代对象。
    func也可以是np.ndarray
    返回 val_array(1D), vec_array(2D,axis=1 is index)
    '''
    # 检查传入参数是否已经是矩阵
    if type(func) == np.ndarray:
        matrix = func
        para_num = 0
    else:
        # func 为tbmodel类的方法，因而参数个数要去掉self
        para_num:int = func.__code__.co_argcount -1
    if para_num == 0:
        try:
            matrix
        except:
            matrix = func()
        eigenvalue1, eigenvector1 = np.linalg.eig(matrix)
        arg = np.argsort(eigenvalue1.real)
        eigenvalue = eigenvalue1[arg]
        eigenvector = eigenvector1[:,arg]
        return  eigenvalue,eigenvector

    # solve matrix, results append to val_array, vec_array
    def calcu(matrix,val_array, vec_array):
        eigenvalue1, eigenvector1 = np.linalg.eig(matrix)
        arg = np.argsort(eigenvalue1.real)
        eigenvalue = eigenvalue1[arg]
        eigenvector = eigenvector1[:,arg]
        val_array = np.append(val_array, eigenvalue)
        vec_array = np.concatenate((vec_array, eigenvector), axis=1)
        return val_array, vec_array

    if para_num == 1:
        n = func(0).shape[0]
        val_array = np.empty(0)*0j
        vec_array = np.empty((n,0))*0j
        for item in data:
            matrix = func(item)
            val_array, vec_array = calcu(matrix, val_array, vec_array)
    if para_num == 2:
        n = func(0,0).shape[0]
        val_array = np.empty(0)*0j
        vec_array = np.empty((n,0))*0j
        for kx, ky in zip(data[0], data[1]):
            matrix = func(kx,ky)
            val_array, vec_array = calcu(matrix, val_array, vec_array)
    if para_num == 3:
        n = func(0,0,0).shape[0]   
        val_array = np.empty(0)*0j
        vec_array = np.empty((n,0))*0j 
        for kx, ky, kz in zip(data[0], data[1], data[2]):
            matrix = func(kx,ky,kz)
            val_array, vec_array = calcu(matrix, val_array, vec_array)
    return  val_array,vec_array



def ray_casting_method(x, y, polygon):
    '''
    射线法确定点(x,y)是否在多边形内
    '''
    n = len(polygon)
    count = 0
    for i in range(n):
        p1 = polygon[i]
        p2 = polygon[(i+1)%n]
        if y < min(p1[1], p2[1]):
            continue
        if y > max(p1[1], p2[1]):
            continue
        # 遍历每条边时，区间前闭后开[p1,p2),避免重复计算
        if y == p2[1]:
            continue
        # 注意保证p2[1]-p1[1]不为0
        xinters = (y-p1[1])*(p2[0]-p1[0])/(p2[1]-p1[1])+p1[0]
        if x < xinters:
            count += 1
        # 如果在边上直接返回True
        if x == xinters:
            count += 1
            return True
    return bool(count % 2)



def cut_OBC(matrix, x_pos, y_pos, polygon):
    '''
    matrix为矩阵，x_pos, y_pos为矩阵对应的坐标，
    polygon为多边形顶点坐标。
    返回多边形内的矩阵和对应的坐标
    射线法确定多边形内的点
    '''
    
    x_pos_new = np.array(x_pos).copy()
    y_pos_new = np.array(y_pos).copy()
    matrix_new = matrix.copy()

    for i in np.arange(len(x_pos)-1, -1, -1):
        if not ray_casting_method(x_pos[i], y_pos[i], polygon):
            x_pos_new = np.delete(x_pos_new, i)
            y_pos_new = np.delete(y_pos_new, i)
            matrix_new = np.delete(matrix_new, i, axis=0)
            matrix_new = np.delete(matrix_new, i, axis=1)

    return matrix_new, x_pos_new, y_pos_new





    
    
   


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='tbcal',
    version='1.0.8',
    packages=find_packages(),
    install_requires=[
        'numpy',
    ],
    author='ShiyinJia',
    author_email='ShiyinJia@foxmail.com',
    description='A package for tight-binding calculations',
    url='https://github.com/Jia-Shiyin/TB_model_calculation',  
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)





import plotly.graph_objects as go
from plotly.subplots import make_subplots
import numpy as np
import ipywidgets as widgets
from ipywidgets import interact, HBox, VBox
from IPython.display import clear_output
import dash
from dash import dcc, html
from dash.dependencies import Output, Input, State


def plot_bands(x, y):
    fig = go.Figure(data = go.Scatter(x=x, y=y, mode='markers',marker_line_width=1))
    fig.show()




def plot_interactive_bands(
        x1, y1, cdata, x2, y2, 
        marker_size1=1, marker_size2=1,
        hermitian: bool = True,
    ):
    '''
    x1(1D), y1(1D)为subplot1(能带图)的的x轴，y轴数据
    z2(2D)为对应subplot1数据点的subplot2(场分布)数据源，axis=1与x1,y1等长
    x2(1D), y2(1D)为对应z2数据的坐标点，与x1,y1等长
    '''
    customdata = cdata.T  # axis=1 is index in tightbounding.tbcalculation
    f = make_subplots(
        rows=1, cols=2,
        column_widths=[0.6, 0.4],
        row_heights=[1],
        specs=[[{"type": "scatter"}, {"type": "scatter"}]],
        subplot_titles=("Energy band", "Field distribution"),
        figure=go.FigureWidget()
    )

    # 显示复数能量，而画图数据只有实部
    text_values = [f"{y1_i:.2f}" for y1_i in y1]
    band = go.Scatter(
        x=x1, y=y1.real, mode='markers', customdata = customdata.real,
        hovertemplate='%{x}, E=%{text}', text = text_values, 
        showlegend=False,
    )
    band.marker = {
        'line_width': 1, 'line_color':'rgba(0,0,0,0.5)',
    }
    field = go.Scatter(
        x=x2, y=y2, mode='markers', customdata = np.ones(x2.shape[0]),
        hovertemplate='%{marker.color:.2f}', showlegend=False,
    ) 
    field.marker = {
        'size': [90*marker_size2/np.power(y2.shape[0],1/2)] * y1.shape[0],
        'line_color':'rgba(0,0,0,0.3)', 'cauto':True,
        'color': customdata[0].real*0, 'cmin':0,
    }
   

    f.add_trace(band ,row=1, col=1)
    f.add_trace(field,row=1, col=2)
    f.update_layout(
        margin=dict(r=20, t=50, b=50, l=20),
        xaxis2=dict(scaleanchor='y2')    # subplot2 x,y等比例
    )
   

    # 单击能带画场分布
    if hermitian:
        c_init = ['lightblue'] * y1.shape[0]
    else:
        f.data[0].marker.cmid = 0
        c_init = y1.imag
    s_init = [8*marker_size1]* y1.shape[0]
    f.data[0].marker.color = c_init
    f.data[0].marker.size = s_init   


    def update_point(trace, points, selector):
        for i in points.point_inds:
            update_point_ind(i)


    def update_point_ind(i):  
        c = c_init.copy()
        s = s_init.copy()
        if hermitian:
            c[i] = 'rgba(0,0,0,0.7)'
        s[i] = 20
        z = f.data[0].customdata[i]        
        with f.batch_update():
            f.data[0].marker.color = c
            f.data[0].marker.size = s
            f.data[1].marker.color = z
            f.data[1].customdata[0] = i
            if normE_Ez_dropdown.value == 'Ez':
                max_min = np.abs(z).max()
                f.data[1].marker.cmax = max_min
                f.data[1].marker.cmin = -max_min
            elif normE_Ez_dropdown.value == 'normE':
                f.data[1].marker.cmax = np.max(z)
                f.data[1].marker.cmin = 0

        


    f.data[0].on_click(update_point)


    # 更新颜色映射
    colorscale_dropdown = widgets.Dropdown(
        options=[ 'hot', 'Rainbow', 'PuOr', 'Blues', 'Reds','Purples','RdBu'],
        value='RdBu',  
        description='Colorscale:'
    )

    
    def update_colorscale(colorscale):
        with f.batch_update():  # 批量更新以提高性能
            f.data[1].marker.colorscale = colorscale
    
    interact(update_colorscale, colorscale=colorscale_dropdown)


    # 更新场分布形式
    normE_Ez_dropdown = widgets.Dropdown(
        options=["normE", "Ez"],
        value="normE", 
        description="Data:"
    )


    def update_subplot_data(selected_data):
        if selected_data == 'Ez':
            f.data[0].customdata = customdata.real
            update_colorscale('RdBu')
        elif selected_data == 'normE':
            f.data[0].customdata = np.abs(customdata)
            update_colorscale('Purples')
        update_point_ind(int(f.data[1].customdata[0]))


    interact(update_subplot_data, selected_data=normE_Ez_dropdown)    


    # 排布下拉列表
    controls = HBox([normE_Ez_dropdown, colorscale_dropdown])
    clear_output(wait=True)  # 防止多次输出
    return VBox([controls, f])




def plot_interactive_bands_dash(
        x1, y1, cdata, x2, y2, z2=None,
        marker_size1=1, marker_size2=1,
        hermitian: bool = True,
        port=8051
    ):
    '''
    浏览器打开 http://127.0.0.1:8051 查看交互图
    port=8051 可以修改端口
    '''

    customdata = cdata.T  # axis=1 is index in tightbounding.tbcalculation

    text_values = [f"{y1_i:.2f}" for y1_i in y1]
    band = go.Figure(
        data = go.Scatter(
            x=x1, y=y1.real, mode='markers', 
            hovertemplate='%{x}, E=%{text}', text = text_values, showlegend=False,
            marker = {
                'line_width': 1, 'line_color':'rgba(0,0,0,0.5)',
                'color': 'lightblue', 'size': 10*marker_size1,
            },
        ),
        layout = {'title': 'Energy band', 'title_font_size': 30,
                  'height': 600, 'width': 800, 
                  'yaxis': {'title':'Energy', 'title_font_size': 25},
        }
    )

    # 判断场分布是 2D or 3D
    if np.all(z2 == None):
        field = go.Figure(
            data = go.Scatter(
                x=x2, y=y2, mode='markers', customdata = np.ones(x2.shape[0]),
                hovertemplate='%{marker.color:.2f}', showlegend=False,
                marker = {
                    'size': [200*marker_size2/np.power(y2.shape[0],1/2)] * y1.shape[0],
                    'line_color':'rgba(0,0,0,0.3)', 'cauto':True,
                    'color': customdata[0].real*0, 'cmin':0, 
                },
            ),
            layout = {'title': 'Field distribution', 'title_font_size': 30,
                    'xaxis': {'scaleanchor':'y','title':'x', 'title_font_size': 25},
                    'yaxis': {'title':'y', 'title_font_size': 25},
                    'height': 600, 'width': 600,
            }
        )
    else:
        field = go.Figure(
            data = go.Scatter3d(
                x=x2, y=y2, z=z2, mode='markers', customdata = np.ones(x2.shape[0]),
                hovertemplate='%{marker.color:.2f}', showlegend=False,
                marker = {
                    'size': [120*marker_size2/np.power(y2.shape[0],1/3)] * y1.shape[0],
                    'line_color':'rgba(0,0,0,0.3)', 'cauto':True,
                    'color': customdata[0].real*0, 'cmin':0, 
                },
            ),
            layout = {'title': 'Field distribution', 'title_font_size': 30,
                  'scene':{'xaxis':{'range':(y2.min()-0.5, y2.max()+0.5),'title':'x', 'title_font_size': 25},
                           'yaxis':{'range':(y2.min()-0.5, y2.max()+0.5),'title':'y', 'title_font_size': 25},
                           'zaxis':{'range':(y2.min()-0.5, y2.max()+0.5),'title':'z', 'title_font_size': 25}},         
                  'height': 600, 'width': 600,
            }
        )
        
    
   
    app = dash.Dash(__name__)
    app.layout = html.Div(children=[

        html.H1("Interactive Band Structure Diagram for TB Model"),
        html.Div([
            dcc.Graph(id='subplot1', figure=band),
            dcc.Graph(id='subplot2', figure=field),
        ], style={'display': 'flex'}),
        html.Div([
            html.Label('Colorscale:'),
            dcc.Dropdown(
                options=[{'label': i, 'value': i} for i in ['hot', 'Rainbow', 'PuOr', 'Blues', 'Reds','Purples','RdBu']],
                value='RdBu',  
                id='colorscale-dropdown',
                style={'margin-left': '10px','margin-right': '30px', 'min-width': '200px'}
            ),
            html.Label('Data:'),
            dcc.Dropdown(
                options=[{'label': i, 'value': i} for i in ["normE", "Ez"]],
                value="Ez", 
                id='normE-Ez-dropdown',
                style={'margin-left': '10px','margin-right': '30px', 'min-width': '200px'}
            ), 
            html.Button(
                'pin', id='pin_button', n_clicks=0,
                style={'margin-left': '200px', 'min-width': '120px','fontSize': '100%',
                       'backgroundColor': 'rgb(30, 112, 33)', 'color':'rgb(255,255,255)',
                       'cursor': 'pointer','borderRadius': '5px',
            },),
            html.Button(
                'clear all', id='clear_button', n_clicks=0,
                style={'margin-left': '50px', 'min-width': '150px','fontSize': '100%',
                       'backgroundColor': 'rgb(148, 36, 26)', 'color':'rgb(255,255,255)',
                       'cursor': 'pointer','borderRadius': '5px',
            },),
        ], style={'display': 'flex', 'fontSize': '150%'} ),
        html.Div(children=[],id='pinned_graphs',
                 style={'scale':'60%','display': 'grid', 
                        'gridTemplateColumns': 'repeat(4, 1fr)','height': '80vh','width': '20vh'}),   
    ], style={'margin-left': '100px', 'margin-top': '50px'})


    # 定义回调函数，点击band更新field
    # 定义回调函数，更新颜色映射和场分布形式
    @app.callback(
        Output('subplot2', 'figure'),
        Output('subplot1', 'figure'),
        Input('subplot1', 'clickData'),  # 监听点击事件
        Input('colorscale-dropdown', 'value'),
        Input('normE-Ez-dropdown', 'value'),
    )

    def update_graph(clickData, colorscale, selected_data):     #, current_band 
        field.data[0].marker.colorscale = colorscale
        if clickData:
            point_index = clickData['points'][0]['pointIndex']
            # 更新subplot2的title
            new_title = f"Field x={x1[point_index]:.2f}, y={y1[point_index]:.2f}"
            field.update_layout(title=new_title)
            # 获取点击点的customdata
            custom_val = customdata[point_index]
            # 更新subplot2
            if selected_data == 'Ez':
                field.data[0].marker.color = custom_val.real
                field.data[0].marker.update({'cmin':None, 'cmax':None, 'cmid':0})
            elif selected_data == 'normE':
                field.data[0].marker.color = np.abs(custom_val)
                field.data[0].marker.update({'cmin':0, 'cmax':None, 'cmid':None})
            # 根据选中的点更新subplot1的颜色
            color_array = ['lightblue'] * len(x1)
            size_array = [10*marker_size1] * len(x1)
            color_array[point_index] = 'rgba(0,0,0,0.7)' 
            size_array[point_index] = 20
            band.data[0].marker.color = color_array
            band.data[0].marker.size = size_array
        # 恢复布局状态
        #band.update_layout(current_band.get('layout'))   # 减慢相应速度
        field.update_layout(title_font_size=30)
        return field, band
    

    # 定义回调函数，pin将当前场分布在下方保持显示，clear all清空下方图片
    @app.callback(
        Output('pinned_graphs', 'children'),
        Output('clear_button', 'n_clicks'),
        Input('pin_button', 'n_clicks'),
        Input('clear_button', 'n_clicks'),
        State('pinned_graphs', 'children'),
        State('subplot2', 'figure'),
    )
    def update_pinned_graph(pin_button, clear_button,  pinned_graphs, field):
        if pin_button:
            pinned_graphs.append(dcc.Graph(id=str(pin_button), figure=field))
        if clear_button:
            pinned_graphs = []
            clear_button = 0
        return pinned_graphs, clear_button

    try:
        app.run_server(debug=True, port=port, jupyter_mode="external")
    except:
        app.run(debug=True, port=port, jupyter_mode="external")
    print('http://127.0.0.1:{:d}'.format(port))

    return app




